/* ascii ICSP implementation for the Arduino NANO
 * (c) Robert Rozee  2015
 *
 * Below is the currently implemented command set:
 *
 * 'd' : TDI = 0, TMS = 0, read_flag = 0	0x64
 * 'e' : TDI = 0, TMS = 1, read_flag = 0
 * 'f' : TDI = 1, TMS = 0, read_flag = 0 
 * 'g' : TDI = 1, TMS = 1, read_flag = 0
 *
 * 'D' : TDI = 0, TMS = 0, read_flag = 1	0x44
 * 'E' : TDI = 0, TMS = 1, read_flag = 1 
 * 'F' : TDI = 1, TMS = 0, read_flag = 1 
 * 'G' : TDI = 1, TMS = 1, read_flag = 1 
 *
 * (If read_flag = 1 then respond with TDO value of '0' or '1')
 *
 * '.' : no operation, used for formatting
 * '>' : request a sync response of '<'
 * '<' : request a sync response of '>'
 *
 * '0' : clock out a 0 on PGD pin
 * '1' : clock out a 1 on PGD pin
 * '-' : clock in single PGD bit	(*** for older device families)
 *
 * '2' : set MCLR low
 * '3' : set MCLR high
 *
 * '4' : turn power to target OFF
 * '5' : turn power to target ON
 *
 * '6' : turn OFF Vpp			(*** for older device families)
 * '7' : turn ON Vpp			(*** for older device families)

 * '8' : insert 10mS delay 
 * '@' : return A0..A5 inputs as 6 lines of text, CRLF terminated
 * '?' : return ID string, "ascii ICSP v1x"
 *
 * note 1: version number is a single numeric digit followed by single UC letter
 *         if backwards compatibility preserved then only letter needs to change
 *         if compatibility is broken then digit should increment, ie 1d -> 2a
 *
 * note 2: 2-wire, 2-phase transaction can be implemented with commands '0' and '1'
 *
 * note 3: commands '-', '6', '7' are intended to possibly allow the programming
 *         of other PIC families that use a more primitave ICSP command set. these
 *         devices are likely to have much less flash storage, so any added time
 *         overhead is not a major concern
 *
 * note 4: analog inputs A0 and A1 should be reserved for reading Vcc and Vpp

Interface pins on Arduino:
-------------------------
PGC    : (D2) open collector output, 3k3 pullup to Vcc (3v3)
PGD    : (D3) open collector output, 3k3 pullup to Vcc (3v3)
MCLR   : (D4) open collector output, optional 3k3 pullup to Vcc (3v3)

Vcc (multiple pins) : fed from multiple 5v output pins via current limiting 
resistors (100r, 17mA ea), with a 3v3 zener diode to ground. alternatively, 
replace zener with 3v3 LDO regulator and make resistor values smaller (22r 
should do)

RST    : base drive for external MCLR switching transistor
Vpp    : drive for external Vpp switch, relay or opto coupler


2-wire, 4-phase transaction:
---------------------------
PGD := TDI
pulse PGC high
PGD := TMS
pulse PGC high
PGD := 1 (hi-Z with 3k3 pullup)
pulse PGC high
TDO := PGD
pulse PGC high


Enter ICSP mode:
---------------
MCLR := 0
PGD := 0
PGC := 0
V3v3 := 1		(apply power to target, wait 50mS to stabilize)
pulse MCLR high
(pause P18)
clock out "MCHP" signature
(pause P19)
MCLR := 1
(pause P7)

command string: "5.88888.32.0100.1101.0100.0011.0100.1000.0101.0000.3.8"


Exit ICSP mode:
--------------
MCLR := 0	(hold target in reset)
Vcc := 0	(target now powered down)

command string: "88888.4"    (first wait 50mS to ensure target is no longer busy)


Using an Arduino Nano just as a USB to serial bridge:
----------------------------------------------------
if pins 28 and 29 are jumpered together (RESET and GND) then the 328P will be 
held in reset with the processors TxD and RxD pins hi-Z. while in this state the 
USB to serial bridge portion of the Nano can be used for communicating with a 
target processor.



Arduino code:
************/


int PGC  = 2;
int PGD  = 3;
int MCLR = 4;
int Vcc1 = 5;
int Vcc2 = 6;
int Vcc3 = 7;
int RST  = 8;
int Vpp  = 9;
int LED = 13;

bool VppON = false;
int LEDxx = 0;


void setup()
{
  digitalWrite(PGC,  LOW);		// PGC,  open collector /w 3k3 pullup
  digitalWrite(PGD,  LOW);		// PGD,  open collector /w 3k3 pullup
  digitalWrite(MCLR, LOW);		// MCLR, open collector /w 3k3 pullup
//digitalWrite(Vcc1, HIGH);		// Vcc1 )
//digitalWrite(Vcc2, HIGH);		// Vcc2 )  hi-Z or +5v
//digitalWrite(Vcc3, HIGH);		// Vcc3 )

// pins all default to inputs, so the above does not change the
// levels on the pins. it does set up D2..D4 as 'open collector'
// and D5..D7 as switched +5v. hereon in we change the state of
// each pin by switching pinMode() between INPUT and OUTPUT.
//
// addendum:
// seems that when switching to input, the output state is not
// preserved. hence the HIGH writes to Vcc pins only works once;
// these writes are now carried out just before each time Vcc is
// turned on. note that this is not an issue when pins are used
// as open-collector (as is the case with PGC, PGD, MCLR).

  digitalWrite(RST, LOW);
  digitalWrite(Vpp, LOW);
  digitalWrite(LED, LOW);
  
  pinMode(RST, OUTPUT);			// not MCLR (to drive base of NPN OC)
  pinMode(Vpp, OUTPUT);			// Vpp enable output (use opto switch)
  pinMode(LED, OUTPUT);			// status LED on arduino

  analogReference(INTERNAL);		// select internal 1.1v reference

  Serial.begin(1000000, SERIAL_8N2);
  /* 38400, 115200, 230400, 256000, 460800, 921600 */
}


void clock(int D)
{
//if (D == 0) pinMode(PGD, OUTPUT);			// PGD = 0
//       else pinMode(PGD, INPUT);			// PGD = hi-Z
//pinMode(PGC, INPUT);					// HIGH (via 3k3 pullup)
//pinMode(PGC, OUTPUT);					// LOW

// below lines use direct port manipulation to improve speed

  if (D == 0) DDRD |= B00001000;			// PGD = 0
         else DDRD &= B11110111;			// PGD = hi-Z
  delayMicroseconds(1);
  DDRD &= B11111011;					// HIGH (via 3k3 pullup)
  delayMicroseconds(1);
  DDRD |= B00000100;					// LOW
  delayMicroseconds(1);
}


void loop()
{

   if (LEDxx == 0) digitalWrite(LED, HIGH);		// turn status LED ON
   if (LEDxx == 1) digitalWrite(LED, LOW);		// turn status LED OFF
   LEDxx = ++LEDxx & 0x000F;
  
  while (Serial.available())				// loop while data in buffer
  {							// (buffer size is 64 bytes)

//  digitalWrite(LED, HIGH);				// (monitor time spent in loop)

    int I = Serial.read();
    switch (char(I))
    {

// 'd','e','f','g': write TDI and TMS, no read back

      case 'd':
        clock(0);					// TDI = 0
        clock(0);					// TMS = 0
        clock(1);					// configure PGD as input
        clock(1);					// finished reading TDO
      break;

      case 'e':
        clock(0);					// TDI = 0
        clock(1);					// TMS = 1
        clock(1);					// configure PGD as input
        clock(1);					// finished reading TDO
      break;

      case 'f':
        clock(1);					// TDI = 1
        clock(0);					// TMS = 0
        clock(1);					// configure PGD as input
        clock(1);					// finished reading TDO
      break;

      case 'g':
        clock(1);					// TDI = 1
        clock(1);					// TMS = 1
        clock(1);					// configure PGD as input
        clock(1);					// finished reading TDO
      break;


// 'D','E','F','G': write TDI and TMS, read back TDO

      case 'D':
        clock(0);					// TDI = 0
        clock(0);					// TMS = 0
        clock(1);					// configure PGD as input
//      Serial.print(digitalRead(PGD) ? '1':'0');	// transmit TDO back to host
        Serial.print(PIND & B00001000 ? '1':'0');	// direct port access for speed
        clock(1);					// finished reading TDO
      break;

      case 'E':
        clock(0);					// TDI = 0
        clock(1);					// TMS = 1
        clock(1);					// configure PGD as input
//      Serial.print(digitalRead(PGD) ? '1':'0');	// transmit TDO back to host
        Serial.print(PIND & B00001000 ? '1':'0');	// direct port access for speed
        clock(1);					// finished reading TDO
      break;

      case 'F':
        clock(1);					// TDI = 1
        clock(0);					// TMS = 0
        clock(1);					// configure PGD as input
//      Serial.print(digitalRead(PGD) ? '1':'0');	// transmit TDO back to host
        Serial.print(PIND & B00001000 ? '1':'0');	// direct port access for speed
        clock(1);					// finished reading TDO
      break;

      case 'G':
        clock(1);					// TDI = 1
        clock(1);					// TMS = 1
        clock(1);					// configure PGD as input
//      Serial.print(digitalRead(PGD) ? '1':'0');	// transmit TDO back to host
        Serial.print(PIND & B00001000 ? '1':'0');	// direct port access for speed
        clock(1);					// finished reading TDO
      break;

// '0','1': used to clock out "MCHP" signature for ICSP entry

      case '0':						// clock out a 0 bit on PGD pin
        clock(0);					// PGD = 0
      break;

      case '1':						// clock out a 1 bit on PGD pin
        clock(1);					// PGD = 1
      break;

      case '-':						// clock in single PGD bit
        clock(1);					// configure PGD as input
//      Serial.print(digitalRead(PGD) ? '1':'0');
        Serial.print(PIND & B00001000 ? '1':'0');
      break;

// '2','3': set MCLR low, clock out signature, set MCLR high

      case '2':						// set MCLR low
        pinMode(MCLR, OUTPUT);				// MCLR = 0
        if (!VppON) digitalWrite(RST, HIGH);		// RST = 1
      break;

      case '3':						// set MCLR high
        pinMode(MCLR, INPUT);				// MCLR = 1
        digitalWrite(RST, LOW);				// RST = 0
      break;

// '4','5': control power supply to target

      case '4':						// turn power to target OFF
        pinMode(MCLR, OUTPUT);				// hold target in reset
	if (!VppON) digitalWrite(RST, HIGH);

        pinMode(Vcc1, INPUT);				// hi-Z
        pinMode(Vcc2, INPUT);				// hi-Z
        pinMode(Vcc3, INPUT);				// hi-Z
//      DDRD = DDRD & B00011111;
      break;

      case '5':						// turn power to target ON
        pinMode(PGC, OUTPUT);				// PGC = 0
        pinMode(PGD, OUTPUT);				// PGD = 0
        pinMode(MCLR, OUTPUT);				// hold target in reset
	if (!VppON) digitalWrite(RST, HIGH);

        digitalWrite(Vcc1, HIGH);			// Vcc1 )
        digitalWrite(Vcc2, HIGH);			// Vcc2 )  kludge to reset to +5v
        digitalWrite(Vcc3, HIGH);			// Vcc3 )
        
        pinMode(Vcc1, OUTPUT);				// +5v
        pinMode(Vcc2, OUTPUT);				// +5v
        pinMode(Vcc3, OUTPUT);				// +5v
//      DDRD = DDRD | B11100000;
      break;

// HV programming commands, for older device families that require Vpp

      case '6':						// turn OFF Vpp
        digitalWrite(Vpp, LOW);				// Vpp = 0
        VppON = false;
      break;

      case '7':						// turn ON Vpp
        digitalWrite(RST, LOW);				// RST = 0 (safety lockout)
        digitalWrite(Vpp, HIGH);			// Vpp = 1
        VppON = true;
      break;

// miscellaneous other commands

      case '8':						// insert 10mS delay 
        delay(10);
      break;

      case '>':						// request a sync response of '<'
        Serial.print('<');
      break;

      case '<':						// request a sync response of '>'
        Serial.print('>');
      break;

      case '.':						// no operation, used for formatting
      break;

      case '@':
        Serial.println(analogRead(A0));
        Serial.println(analogRead(A1));
        Serial.println(analogRead(A2));
        Serial.println(analogRead(A3));
        Serial.println(analogRead(A4));
        Serial.println(analogRead(A5));
      break;
      
      case '?':						// return ID string, "ascii ICSP v1"
        Serial.print("ascii ICSP v1B");
      break;
    }	// end of switch
//  digitalWrite(LED, LOW);     			// (monitor time spent in loop)
  }	// end of while
}	// end of function loop()




//  pinMode(pin, OUTPUT);				// drive pin to set value
//  pinMode(pin, INPUT);				// hi-Z state
