// last edited: Monday, 1 June, 2015 (BETA 05)

/*
 * Interface to PIC32 ICSP port using bitbang adapter.
 *
 * Copyright (C) 2014 Serge Vakulenko
 *
 * Additions for talking to bitbang programmer Copyright (C) 2015 Robert Rozee
 *
 * This file is part of PIC32PROG project, which is distributed
 * under the terms of the GNU General Public License (GPL).
 * See the accompanying file "COPYING" for more details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include <windows.h>

#include "adapter.h"
#include "pic32.h"

typedef struct {
    /* Common part */
    adapter_t adapter;
    const char *name;                // put the name of the comm port in here, eg "COM3"

// ***********************************************************************************
    char *path;                      // port name fully qualified for windows
    HANDLE devfd;                    // serial port handle
    int BitsToRead;                  // number of 'bits' waiting in Rx buffer
    unsigned TotalBitPairsSent;      // count of total # of TDI and TMS pairs sent
    unsigned TotalBitsReceived;      // count of total # of TDO bits recieved
    unsigned MaxBufferedWrites;      // max continuous string of writes before read
    unsigned RunningWriteCount;      // a running count of writes, reset by read
    int PendingHandshake;            // indicates a return handshake is expected
// ***********************************************************************************

    unsigned use_executive;
    unsigned serial_execution_mode;
} bitbang_adapter_t;

int DBG = 0;                      // extra internal low-level diagnostics flag


/*
 * Calculate checksum.
 */
static unsigned calculate_crc (unsigned crc, unsigned char *data, unsigned nbytes)
{
    static const unsigned short crc_table [16] = {
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
        0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    };
    unsigned i;

    while (nbytes--) {
        i = (crc >> 12) ^ (*data >> 4);
        crc = crc_table[i & 0x0F] ^ (crc << 4);
        i = (crc >> 12) ^ (*data >> 0);
        crc = crc_table[i & 0x0F] ^ (crc << 4);
        data++;
    }
    return crc & 0xffff;
}



/*
 * Sends a command ('8')to the programmer telling it to insert
 * a 10mS delay in the datastream being sent to the target. This
 * is the only reliable way to create a delay at the target.
 * (by RR)
 */ 
static void bitbang_delay10mS (bitbang_adapter_t *a)
{
  int n;
  char ch;

  ch='8';
  WriteFile (a->devfd, &ch, 1, (LPDWORD)((void *)&n), NULL);
}



/*
 * Current version of bitbang_send, sends a string of data out to the target encoded
 * as ASCII characters to be interpreted by an intellenent ICSP programmer.

 * NOTE: this version of pic32prog implements the 2-wire communications mode via the
 * ICSP port. This means we need to encode TDI, TMS, and read_flag into one character,
 * sending all three out at once to an Arduino NANO acting as an intellige programmer.
 * If read_flag is set, then the programmer needs to respond with a single character
 * indicating the value of TDO that the target responds with.
 *
 * Below is the currently implemented command set:
 *
 * 'd' : TDI = 0, TMS = 0, read_flag = 0   0x64
 * 'e' : TDI = 0, TMS = 1, read_flag = 0
 * 'f' : TDI = 1, TMS = 0, read_flag = 0 
 * 'g' : TDI = 1, TMS = 1, read_flag = 0
 *
 * 'D' : TDI = 0, TMS = 0, read_flag = 1   0x44
 * 'E' : TDI = 0, TMS = 1, read_flag = 1 
 * 'F' : TDI = 1, TMS = 0, read_flag = 1 
 * 'G' : TDI = 1, TMS = 1, read_flag = 1 
 *
 * '.' : no operation, used for formatting
 * '>' : request sync response - '<'
 *
 * '0' : clock out a 0 on PGD pin
 * '1' : clock out a 1 on PGD pin

 * '2' : set MCLR low
 * '3' : set MCLR high
 * '4' : turn target power off
 * '5' : turn target power on
 * '8' : insert 10mS delay 
 * '?' : return ID string, "ascii JTAG XN
 *
 * if the request is 'D'..'E', then respond with '0'/'1' to indicate TDO = 0/1
 *
 * (by RR)
 */
static void bitbang_send (bitbang_adapter_t *a,
    unsigned tms_nbits, unsigned tms,
    unsigned tdi_nbits, unsigned long long tdi, int read_flag)
{
// ************************************************************************************
// TMS is a command of 0 to 14 bits length, sent LSB first. TDI = 0 throughout.
//
// if nTDI<>0 then send TMS = 1-0-0 (TDI = 0).
//
// Next we send out the TDI bits (up to 64, LSB first), with TMS=0 for n-1 bits;
// the last TDI bit should be accompanied with TMS = 1.
//
// if nTDI<>0 then send TMS = 1-0 (TDI = 0).
// 
// NOTE: if read_flag set, then read TDO on each TDI bit.
//
// ************************************************************************************

  char buffer[110];                    // @@@@@@@@@@ BUFFERED WRITES VERSION @@@@@@@@@@
  int index;
  int i, n;
  char ch;

  index = 0;

  for (i = tms_nbits; i > 0; --i) {              // for each of the n bits...
    ch = ((tms & 1) + 0x64);                     // d, e, f, g           
    buffer [index++] = ch;                       // append to buffer
    tms >>= 1;                                   // shift TMS right one bit
  }

  if (DBG && (tms_nbits != 0)) buffer[index++] = '.';   // spacer, ignored by programmer

  if (tdi_nbits != 0) {                          // 1-0-0 if nTDI <> 0
    ch = (1 + 0x64);
    buffer[index++] = ch;
    ch = (0 + 0x64);
    buffer[index++] = ch;
    ch = (0 + 0x64);
    buffer[index++] = ch;
    if (DBG) buffer[index++] = '.';              // spacer, ignored by programmer
  }

// fixup required when changing from JTAG to ICSP - TDO bits appear one slot earlier
  if (read_flag) { buffer[index - (DBG ? 2 : 1)] ^= 0x20; }

  for (i = tdi_nbits; i > 0; --i) {
    ch = ((tdi & 1) << 1) + (i==1) +             // TMS=0 for n-1 bits, then 1 on last bit     
         (read_flag ? 0x44 : 0x64);              // UC = read, LC = no read
    buffer[index++] = ch;                        // append to buffer
    tdi >>= 1;                                   // shift TDI right one bit
  }

// fixup required when changing from JTAG to ICSP - TDO bits appear one slot earlier
  if (read_flag) { buffer[index - 1] ^= 0x20; }

  if (tdi_nbits != 0) {                          // 1-0 if nTDI <> 0
    if (DBG) buffer[index++] = '.';              // spacer, ignored by programmer
    ch = (1 + 0x64);
    buffer[index++] = ch;
    ch = (0 + 0x64);
    buffer[index++] = ch;
  }


/////////////////////////////////////////////////////////////////////////////////////////////////
/////////// control handshaking for JTAG programmers with small (64 byte) Rx buffers ////////////
///////////////////////////////////////////////////////////////////////////////////////////////
  if (a->PendingHandshake) {
    if (a->RunningWriteCount > a->MaxBufferedWrites) a->MaxBufferedWrites = a->RunningWriteCount;
    a->RunningWriteCount = 0;
    a->PendingHandshake = 0;

    n = 0;
    while (n == 0) ReadFile (a->devfd, &ch, 1, (LPDWORD)((void *)&n), NULL);
    if (ch != '<') fprintf(stderr, "WARNING - handshake error (BB send)!!\n");
  }
  if (!read_flag) {
    buffer[index++] = '>';
    a->PendingHandshake = 1;
  }

////////////////////////////////////////////////////////////////////////////////////////////////

  buffer[index] = 0;                             // append trailing zero so can print as a string
  if (DBG) fprintf (stderr, "n=%i, <%s> read=%i\n", index, buffer, read_flag);

  a->TotalBitPairsSent += index;
  a->RunningWriteCount += index;

  if (a->BitsToRead != 0) fprintf(stderr, "WARNING - write while pending read!!\n");

  WriteFile (a->devfd, &buffer, index, (LPDWORD)((void *)&n), NULL);

  if (n != index) fprintf(stderr, "WARNING - not all bits written (BB send)!!\n");
  if (read_flag) a->BitsToRead += tdi_nbits;
}



/*
 * (by RR)
 */
static unsigned long long bitbang_recv (bitbang_adapter_t *a)
{
  char buffer[70];
  unsigned long long word;
  int n, i;

/////////////// this code is also duplicated in bitbang_send //////////////////////////////////
  if (a->RunningWriteCount > a->MaxBufferedWrites) a->MaxBufferedWrites = a->RunningWriteCount;
  a->RunningWriteCount = 0;
///////////////////////////////////////////////////////////////////////////////////////////////

  ReadFile (a->devfd, &buffer, a->BitsToRead, (LPDWORD)((void *)&n), NULL);
  if (n != a->BitsToRead) fprintf(stderr, "WARNING - fewer bits read than expected!! %i %i\n", a->BitsToRead, n);

  word = 0;

  for (i = 0; i < n; ++i) {
    if (buffer[i] == '0') buffer[i] = 0;
    if (buffer[i] == '1') buffer[i] = 1;
    if (buffer[i] > 1) fprintf(stderr, "WARNING - unexpected character returned!!\n");
    word |= buffer[i] << i;
  }

  unsigned L4 = word >> 48;
  unsigned L3 = (word >> 32) & 0xFFFF;
  unsigned L2 = (word >> 16) & 0xFFFF;
  unsigned L1 = word & 0xFFFF;

  if (DBG) fprintf (stderr, "TDO = %04x %04x %04x %04x (%i bits)   %I64u (dec)\n", 
                                    L4,  L3,  L2,  L1, a->BitsToRead, word);  

  a->TotalBitsReceived += a->BitsToRead;
  a->BitsToRead = 0;
  return word;
}



/*
 * (by RR)
 */
static void bitbang_ICSP_enable (bitbang_adapter_t *a, int ICSP_EN)
{
  int n;

  if (ICSP_EN)
  {
    char buffer[64] = "5.88888.32.0100.1101.0100.0011.0100.1000.0101.0000.3.8..........";
//                     0000000001111111111222222222233333333334444444444555555555566666
//                     1234567890123456789012345678901234567890123456789012345678901234
// 50mS delay after powerup, pulse MCLR high, send signature, set MCLR high, 10mS delay

    WriteFile (a->devfd, &buffer, 64, (LPDWORD)((void *)&n), NULL);
    if (n != 64) fprintf(stderr, "WARNING - not all bits written (enter ICSP)!!\n");

    mdelay(100);                  // 100mS delay to allow the above to percolate through the system
  }
  else
  {
    char buffer[16] = "88888.4.........";
//                     0000000000111111
//                     1234567890123456
// 50mS delay then cut power
    WriteFile (a->devfd, &buffer, 16, (LPDWORD)((void *)&n), NULL);
    if (n != 16) fprintf(stderr, "WARNING - not all bits written (exit ICSP)!!\n");
  }
}



static void bitbang_close (adapter_t *adapter, int power_on)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;

    /* Clear EJTAGBOOT mode. */
    bitbang_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */
    bitbang_send (a, 6, 31, 0, 0, 0);             /* TMS 1-1-1-1-1-0 */   

    bitbang_ICSP_enable (a, 0);

    printf ("total TDI/TMS pairs sent = %i pairs\n", a->TotalBitPairsSent);
    printf ("total TDO bits received  = %i bits\n",  a->TotalBitsReceived);
    printf ("maximum continuous write = %i chars\n", a->MaxBufferedWrites);

    CloseHandle (a->devfd);            // at this point we are exiting application???
//  free (a);                          // suspect this was causing XP CRASHES, shouldn't be needed anyway
}



/*
 * Read the Device Identification code
 */
static unsigned bitbang_get_idcode (adapter_t *adapter)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;
    unsigned idcode;

    /* Reset the JTAG TAP controller: TMS 1-1-1-1-1-0.
     * After reset, the IDCODE register is always selected.
     * Read out 32 bits of data. */
    bitbang_send (a, 6, 31, 32, 0, 1);
    idcode = bitbang_recv (a);
    return idcode;
}



/*
 * Put device to serial execution mode. This is an alternative version
 * taken directly from the microchip application note. The original 
 * version threw up a status error and then did an exit(-1), but when 
 * the exit was commented out still seemed to function.
 * SEE: "PIC32 Flash Programming Specification 60001145N.pdf"
 * (by RR)
 */
static void serial_execution (bitbang_adapter_t *a)
{
    if (a->serial_execution_mode) {
        return;
    }

    a->serial_execution_mode = 1;

    /* Enter serial execution. */
    if (debug_level > 0)
        fprintf (stderr, "%s: enter serial execution\n", a->name);

    bitbang_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */  // missing M: MTAP     // 1.
    bitbang_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */                         // 2.
    bitbang_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */                            // 3.
    unsigned status = bitbang_recv (a);
    if (debug_level > 0)
        fprintf (stderr, "%s: status %04x\n", a->name, status);
    if ((status & MCHP_STATUS_CPS) == 0) {
        fprintf (stderr, "%s: invalid status = %04x (code protection)\n", a->name, status);   // 4.
        exit (-1);
    }

    bitbang_send (a, 0, 0, 8, MCHP_ASSERT_RST, 0);  /* Xfer data. */                          // 5.

    bitbang_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */  // missing M: MTAP     // 6.
    bitbang_send (a, 1, 1, 5, ETAP_EJTAGBOOT, 0); /* Send command. */                         // 7.


    bitbang_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */  // missing M: MTAP     // 8.
    bitbang_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */                         // 9.
    bitbang_send (a, 0, 0, 8, MCHP_DEASSERT_RST, 0);  /* Xfer data. */    // missing _: "DE_" // 10.
    bitbang_send (a, 0, 0, 8, MCHP_FLASH_ENABLE, 0);  /* Xfer data. */                        // 11.
// comment about the above line: "This command requires a NOP to complete."
    bitbang_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */                         // 12.
}



// *********************************************************************
//
// Shouldn't this check the value of PrAcc returned in the LSB?
// We don't seem to be even reading this back.
// UPDATE: The check is not needed in 4-wire JTAG mode, seems
// to only be required if in 2-wire ICSP mode. It would also
// slow us down horribly.
//
// *********************************************************************
static void xfer_fastdata (bitbang_adapter_t *a, unsigned word)
{
    bitbang_send (a, 0, 0, 33, (unsigned long long) word << 1, 0);
}



static void xfer_instruction (bitbang_adapter_t *a, unsigned instruction)
{
    unsigned ctl;

    if (debug_level > 1)
        fprintf (stderr, "%s: xfer instruction %08x\n", a->name, instruction);

    // Select Control Register
    bitbang_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */

    // Wait until CPU is ready
    // Check if Processor Access bit (bit 18) is set
    do {
        bitbang_send (a, 0, 0, 32, CONTROL_PRACC |    /* Xfer data. */
                                  CONTROL_PROBEN |
                                CONTROL_PROBTRAP |
                                CONTROL_EJTAGBRK, 1);
        ctl = bitbang_recv (a);
    } while (! (ctl & CONTROL_PRACC));

    // Select Data Register
    // Send the instruction
    bitbang_send (a, 1, 1, 5, ETAP_DATA, 0);          /* Send command. */
    bitbang_send (a, 0, 0, 32, instruction, 0);       /* Send data. */

    // Tell CPU to execute instruction
    bitbang_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */
    bitbang_send (a, 0, 0, 32, CONTROL_PROBEN |       /* Send data. */
                             CONTROL_PROBTRAP, 0);
}



static unsigned get_pe_response (bitbang_adapter_t *a)
{
    unsigned ctl, response;

    // Select Control Register
    bitbang_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */

    // Wait until CPU is ready
    // Check if Processor Access bit (bit 18) is set

    int i = 0;
    do {
        bitbang_send (a, 0, 0, 32, CONTROL_PRACC |    /* Xfer data. */
                                  CONTROL_PROBEN |
                                CONTROL_PROBTRAP |
                                 CONTROL_EJTAGBRK, 1);
        ctl = bitbang_recv (a);
        i++;
    } while ( (! (ctl & CONTROL_PRACC)) & (i<100) );

    if (i==100) {
      fprintf (stderr, "%s: PE response, Processor Access bit not set\n", a->name);
      exit (-1);
    }
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////// we have on occasion got stuck in the above loop. in normal circumstances, should not loop more than once ///////
////// (added crude fix to terminate after 100 iterations)                                                      ///////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // Select Data Register
    // Send the instruction
    bitbang_send (a, 1, 1, 5, ETAP_DATA, 0);          /* Send command. */
    bitbang_send (a, 0, 0, 32, 0, 1);                 /* Get data. */
    response = bitbang_recv (a);

    // Tell CPU to execute NOP instruction
    bitbang_send (a, 1, 1, 5, ETAP_CONTROL, 0);       /* Send command. */
    bitbang_send (a, 0, 0, 32, CONTROL_PROBEN |       /* Send data. */
                             CONTROL_PROBTRAP, 0);
    if (debug_level > 1)
        fprintf (stderr, "%s: get PE response %08x\n", a->name, response);
    return response;
}



/*
 * Read a word from memory (without PE).
 */
static unsigned bitbang_read_word (adapter_t *adapter, unsigned addr)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;
    unsigned addr_lo = addr & 0xFFFF;
    unsigned addr_hi = (addr >> 16) & 0xFFFF;

    serial_execution (a);

    //fprintf (stderr, "%s: read word from %08x\n", a->name, addr);
    xfer_instruction (a, 0x3c04bf80);           // lui s3, 0xFF20
    xfer_instruction (a, 0x3c080000 | addr_hi); // lui t0, addr_hi
    xfer_instruction (a, 0x35080000 | addr_lo); // ori t0, addr_lo
    xfer_instruction (a, 0x8d090000);           // lw t1, 0(t0)
    xfer_instruction (a, 0xae690000);           // sw t1, 0(s3)

    bitbang_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// @@@@@@@@@ surely the following line should be using xfer_fastdata ??? @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    bitbang_send (a, 0, 0, 33, 0, 1);             /* Get fastdata. */
    unsigned word = bitbang_recv (a) >> 1;

    if (debug_level > 0)
        fprintf (stderr, "%s: read word at %08x -> %08x\n", a->name, addr, word);
    return word;
}



/*
 * Read a memory block.
 */
static void bitbang_read_data (adapter_t *adapter,
    unsigned addr, unsigned nwords, unsigned *data)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;
    unsigned words_read, i;

    //fprintf (stderr, "%s: read %d bytes from %08x\n", a->name, nwords*4, addr);
    if (! a->use_executive) {
        /* Without PE. */
        for (; nwords > 0; nwords--) {
            *data++ = bitbang_read_word (adapter, addr);
            addr += 4;
        }
        return;
    }

    /* Use PE to read memory. */
    for (words_read = 0; words_read < nwords; words_read += 32) {

        bitbang_send (a, 1, 1, 5, ETAP_FASTDATA, 0);
        xfer_fastdata (a, PE_READ << 16 | 32);      /* Read 32 words */
        xfer_fastdata (a, addr);                    /* Address */

        unsigned response = get_pe_response (a);    /* Get response */
        if (response != PE_READ << 16) {
            fprintf (stderr, "%s: bad READ response = %08x, expected %08x\n",
                a->name, response, PE_READ << 16);
            exit (-1);
        }
        for (i=0; i<32; i++) {
            *data++ = get_pe_response (a);          /* Get data */
        }
        addr += 32*4;
    }
}



/*
 * Download programming executive (PE).
 */
static void bitbang_load_executive (adapter_t *adapter,
    const unsigned *pe, unsigned nwords, unsigned pe_version)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;

    a->use_executive = 1;
    serial_execution (a);

    if (debug_level > 0)
        fprintf (stderr, "%s: download PE loader\n", a->name);

    printf ("   Loading PE: ");

    /* Step 1. */
    xfer_instruction (a, 0x3c04bf88);   // lui a0, 0xbf88
    xfer_instruction (a, 0x34842000);   // ori a0, 0x2000 - address of BMXCON
    xfer_instruction (a, 0x3c05001f);   // lui a1, 0x1f
    xfer_instruction (a, 0x34a50040);   // ori a1, 0x40   - a1 has 001f0040
    xfer_instruction (a, 0xac850000);   // sw  a1, 0(a0)  - BMXCON initialized
printf ("1");
    /* Step 2. */
    xfer_instruction (a, 0x34050800);   // li  a1, 0x800  - a1 has 00000800
    xfer_instruction (a, 0xac850010);   // sw  a1, 16(a0) - BMXDKPBA initialized
printf (" 2");
    /* Step 3. */
    xfer_instruction (a, 0x8c850040);   // lw  a1, 64(a0) - load BMXDMSZ
    xfer_instruction (a, 0xac850020);   // sw  a1, 32(a0) - BMXDUDBA initialized
    xfer_instruction (a, 0xac850030);   // sw  a1, 48(a0) - BMXDUPBA initialized
printf (" 3");
    /* Step 4. */
    xfer_instruction (a, 0x3c04a000);   // lui a0, 0xa000
    xfer_instruction (a, 0x34840800);   // ori a0, 0x800  - a0 has a0000800
printf (" 4 (LDR)");
    /* Download the PE loader. */
    int i;
    for (i=0; i<PIC32_PE_LOADER_LEN; i+=2) {
        /* Step 5. */
        unsigned opcode1 = 0x3c060000 | pic32_pe_loader[i];
        unsigned opcode2 = 0x34c60000 | pic32_pe_loader[i+1];

        xfer_instruction (a, opcode1);      // lui a2, PE_loader_hi++
        xfer_instruction (a, opcode2);      // ori a2, PE_loader_lo++
        xfer_instruction (a, 0xac860000);   // sw  a2, 0(a0)
        xfer_instruction (a, 0x24840004);   // addiu a0, 4

// @@@@@@@@@@@@@@@@@@@@@@ rendered unnecessary by handshaking now implemented in bitbang_write @@@@@@@@@@@@@@@@@@@@@
//      if (((i & 7) == 0) && (a->RunLength == 2))
//          bitbang_synchronize (a);    // ********** needed to ensure we don't overrun the programmers input buffer 
    }
printf (" 5");
    /* Jump to PE loader (step 6). */
    xfer_instruction (a, 0x3c19a000);   // lui t9, 0xa000
    xfer_instruction (a, 0x37390800);   // ori t9, 0x800  - t9 has a0000800
    xfer_instruction (a, 0x03200008);   // jr  t9
    xfer_instruction (a, 0x00000000);   // nop
printf (" 6");
    /* Switch from serial to fast execution mode. */
    bitbang_send (a, 1, 1, 5, TAP_SW_ETAP, 0);
    bitbang_send (a, 6, 31, 0, 0, 0);             /* TMS 1-1-1-1-1-0 */

    /* Send parameters for the loader (step 7-A).
     * PE_ADDRESS = 0xA000_0900,
     * PE_SIZE */
    bitbang_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, 0xa0000900);
    xfer_fastdata (a, nwords);
printf (" 7a (PE)");
    /* Download the PE itself (step 7-B). */
    if (debug_level > 0)
        fprintf (stderr, "%s: download PE\n", a->name);
    for (i=0; i<nwords; i++) {
        xfer_fastdata (a, *pe++);

// @@@@@@@@@@@@@@@@@@@@@@ rendered unnecessary by handshaking now implemented in bitbang_write @@@@@@@@@@@@@@@@@@@@@
//      if (((i & 7) == 0) && (a->RunLength == 2))
//          bitbang_synchronize (a);    // ********** needed to ensure we don't overrun the programmers input buffer 
    }
//    mdelay (10);
    bitbang_delay10mS (a);
printf (" 7b");
    /* Download the PE instructions. */
    xfer_fastdata (a, 0);                       /* Step 8 - jump to PE. */
    xfer_fastdata (a, 0xDEAD0000);

    bitbang_delay10mS (a);

    xfer_fastdata (a, PE_EXEC_VERSION << 16);

    unsigned version = get_pe_response (a);
    if (version != (PE_EXEC_VERSION << 16 | pe_version)) {
        fprintf (stderr, "%s: bad PE version = %08x, expected %08x\n",
            a->name, version, PE_EXEC_VERSION << 16 | pe_version);
        exit (-1);
    }

    printf (" done\n");

    if (debug_level > 0)
        fprintf (stderr, "%s: PE version = %04x\n",
            a->name, version & 0xffff);
}



/*
 * Erase all flash memory.
 */
static void bitbang_erase_chip (adapter_t *adapter)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;

    bitbang_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
    bitbang_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
    bitbang_send (a, 0, 0, 8, MCHP_ERASE, 0);     /* Xfer data. */

//  XferData (MCHP_DE_ASSERT_RST), PIC32MZ devices only.

    int i = 0;
    unsigned status;
    do {
      bitbang_delay10mS (a);
      bitbang_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */
      status = bitbang_recv (a);
      i++;
    } while ( ((status & (MCHP_STATUS_CFGRDY | 
                          MCHP_STATUS_FCBUSY)) != MCHP_STATUS_CFGRDY) 
              & (i<100) );

    if (i==100) {
      fprintf (stderr, "%s: [BB_7] invalid status = %04x (reset)\n", a->name, status);
      exit (-1);
   }      

    /* Leave it in ETAP mode. */
    bitbang_send (a, 1, 1, 5, TAP_SW_ETAP, 0);    /* Send command. */
}



/*
 * Write a word to flash memory. (only seems to be used to write the four configuration words)
 */
static void bitbang_program_word (adapter_t *adapter,
    unsigned addr, unsigned word)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;

    if (debug_level > 0)
        fprintf (stderr, "%s: program word at %08x: %08x\n", a->name, addr, word);
    if (! a->use_executive) {
        /* Without PE. */
        fprintf (stderr, "%s: slow flash write not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to write flash memory. */
    bitbang_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_WORD_PROGRAM << 16 | 2);
    xfer_fastdata (a, addr);                    /* Send address. */
    xfer_fastdata (a, word);                    /* Send word. */

    unsigned response = get_pe_response (a);
    if (response != (PE_WORD_PROGRAM << 16)) {
        fprintf (stderr, "%s: failed to program word %08x at %08x, reply = %08x\n",
            a->name, word, addr, response);
        exit (-1);
    }
}



/*
 * Flash write row of memory.
 */
static void bitbang_program_row (adapter_t *adapter, unsigned addr,
    unsigned *data, unsigned words_per_row)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;
    int i;

    if (debug_level > 0)
        fprintf (stderr, "%s: row program %u words at %08x\n",
            a->name, words_per_row, addr);
    if (! a->use_executive) {
        /* Without PE. */
        fprintf (stderr, "%s: slow flash write not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to write flash memory. */
    bitbang_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_ROW_PROGRAM << 16 | words_per_row);
    xfer_fastdata (a, addr);                    /* Send address. */

//////////////////////////////////////////////////////////////////////
//  the below for loop seems to be where our biggest bottleneck is. //
//////////////////////////////////////////////////////////////////////

    /* Download data. */
    for (i = 0; i < words_per_row; i++) {
        xfer_fastdata (a, *data++);             /* Send word. */
    }

    unsigned response = get_pe_response (a);
    if (response != (PE_ROW_PROGRAM << 16)) {
        fprintf (stderr, "%s: failed to program row at %08x, reply = %08x\n",
            a->name, addr, response);
        exit (-1);
    }
}



/*
 * Verify a block of memory.
 */
static void bitbang_verify_data (adapter_t *adapter,
    unsigned addr, unsigned nwords, unsigned *data)
{
    bitbang_adapter_t *a = (bitbang_adapter_t*) adapter;
    unsigned data_crc, flash_crc;

    if (! a->use_executive) {
        /* Without PE. */
        fprintf (stderr, "%s: slow verify not implemented yet.\n", a->name);
        exit (-1);
    }

    /* Use PE to get CRC of flash memory. */
    bitbang_send (a, 1, 1, 5, ETAP_FASTDATA, 0);  /* Send command. */
    xfer_fastdata (a, PE_GET_CRC << 16);
    xfer_fastdata (a, addr);                    /* Send address. */
    xfer_fastdata (a, nwords * 4);              /* Send length. */

    unsigned response = get_pe_response (a);
    if (response != (PE_GET_CRC << 16)) {
        fprintf (stderr, "%s: failed to verify %d words at %08x, reply = %08x\n",
            a->name, nwords, addr, response);
        exit (-1);
    }

    flash_crc = get_pe_response (a) & 0xffff;

    data_crc = calculate_crc (0xffff, (unsigned char*) data, nwords * 4);
    if (flash_crc != data_crc) {
        fprintf (stderr, "%s: checksum failed at %08x: sum=%04x, expected=%04x\n",
            a->name, addr, flash_crc, data_crc);
        //exit (-1);
    }
}



/*
 * Initialize bitbang adapter.
 * Return a pointer to a data structure, allocated dynamically.
 * When adapter not found, return 0.
 */
adapter_t *adapter_open_bitbang (const char *port)
{
    bitbang_adapter_t *a;

    a = calloc (1, sizeof (*a));
    if (! a) {
        fprintf (stderr, "adapter_open_bitbang: out of memory\n");
        return 0;
    }

    // TODO: find and open the connection to serial device `port'.
    // a->devfd = 0;


// ***************************************************************************
// ******** The following block of code is very windows-specific *************
// **(by RR)******************************************************************

    a->name = port;                            // a string containing "COMn"
    a->path = malloc (strlen(port) + 16);
        
    strcpy(a->path, "\\\\.\\");
    strcat(a->path, port);

    a->devfd = CreateFileA (a->path,
               GENERIC_READ | GENERIC_WRITE,
               0,
               NULL,
               OPEN_EXISTING,
               0,
               NULL);

    if (a->devfd == INVALID_HANDLE_VALUE) {
        fprintf (stderr, "%s: Selected serial communications port not found\n", port);
        free (a);
        return 0;
    }

    DCB settings;

    memset(&settings, 0, sizeof (settings));
    settings.DCBlength = sizeof (settings);

//  38400, 115200, 230400, 256000, 460800, 921600

    if (!BuildCommDCB ("baud=1000000 data=8 parity=N stop=2 xon=off dtr=off rts=off odsr=off octs=off", &settings)) {
      fprintf (stderr, "Failed to successfully populate device control block\n");
      CloseHandle (a->devfd);
      free (a);
      return 0;
    }

    if (!SetCommState (a->devfd, &settings)) {
        fprintf (stderr, "%s: Unable to configure serial communications port\n", port);
        CloseHandle (a->devfd);
        free (a);
        return 0;
    }

    COMMTIMEOUTS timeouts;
//  SERIAL_TIMEOUTS timeouts;

    timeouts.ReadIntervalTimeout         = 250;          // 250mS timeout on read operations
    timeouts.ReadTotalTimeoutMultiplier  = 0;
    timeouts.ReadTotalTimeoutConstant    = 250;
    timeouts.WriteTotalTimeoutMultiplier = 0;
    timeouts.WriteTotalTimeoutConstant   = 250;

    if(!SetCommTimeouts(a->devfd, &timeouts)) {
      fprintf (stderr, "%s: Unable to set time-outs for serial communications port\n", port);
      CloseHandle (a->devfd);
      free (a);
      return 0;
    }

    printf ("\nSerial communications port initialized successfully\n");
    printf (   "*** BETA 05 --- BETA 05 --- BETA 05 --- BETA 05 ***\n");
    printf (   "(for \"bitbang\" support please contact Robert Rozee)\n\n");

// ***************************************************************************
// ******** this is the end of the windows-specific stuff ********************
// ***************************************************************************
    
    printf("   ascii ICSP: ");
    int i, n, ch;
    for (i = 0; i < 20; i++ ) {
     printf(".");
      ch = '>';
      WriteFile (a->devfd, &ch, 1, (LPDWORD)((void *)&n), NULL);
      ReadFile (a->devfd, &ch, 1, (LPDWORD)((void *)&n), NULL);      
      if ((n != 0) && (ch == '<')) { printf("OK1 "); i = 1000; }
    }

    if (i < 1000) {
      fprintf (stderr, "\nNo response from 'ascii ICSP' adapter\n");
      CloseHandle (a->devfd);
      free (a);
      return 0;
    }

    ch='?';
    char buffer1 [15] = "ascii ICSP v1B\0";
    char buffer2 [15] = "..............\0";
    WriteFile (a->devfd, &ch, 1, (LPDWORD)((void *)&n), NULL);
    ReadFile (a->devfd, &buffer2, 14, (LPDWORD)((void *)&n), NULL);
 
    if ((memcmp(buffer1, buffer2, 13) == 0) && (n == 14))
      printf("OK2 (v1%c -> v1%c)\n", buffer1[13], buffer2[13]); 
    else {
      fprintf (stderr, "\nBad response from 'ascii ICSP' adapter\n");
      CloseHandle (a->devfd);
      free (a);
      return 0;
    }

// ***************************************************************************
// ******** this is the end of 'ascii ICSP' ID probe *************************
// ***************************************************************************

    a->BitsToRead = 0;
    a->TotalBitPairsSent = 0;              // count of total # of TDI+TMS bits sent
    a->TotalBitsReceived = 0;              // count of total # of TDO bits recieved
    a->MaxBufferedWrites = 0;              // maximum continuous write length (chars)
    a->RunningWriteCount = 0;              // running count of writes, reset by read
    a->PendingHandshake = 0;

    a->use_executive = 0;
    a->serial_execution_mode = 0;


///////////////////////////////////////////////////////////////////////////////////////
// it is at this point that we start talking to the target. new strategy is to
// keep MCLR low until we need to release it in the routine serial_execution
///////////////////////////////////////////////////////////////////////////////////////


    bitbang_ICSP_enable (a, 1);

// ////////////////////////////////////////////////////////////////////
// the below few lines can be enabled for a forced "emergency" erase //
// ////////////////////////////////////////////////////////////////////
    if (0) {
      bitbang_send (a, 6, 31, 32, 0, 1);    // a, # tms bits, TMS, # tdi bits, TDI, read_flag
      bitbang_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
      bitbang_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
      bitbang_send (a, 0, 0, 8, MCHP_ERASE, 0);     /* Xfer data. */
      bitbang_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */
      bitbang_recv (a);
    }


    /* Reset the JTAG TAP controller: TMS 1-1-1-1-1-0.
     * After reset, the IDCODE register is always selected.
     * Read out 32 bits of data. */
    unsigned idcode;
    bitbang_send (a, 6, 31, 32, 0, 1);    // a, # tms bits, TMS, # tdi bits, TDI, read_flag
    idcode = bitbang_recv (a);
    if ((idcode & 0xfff) != 0x053) {
        /* Microchip vendor ID is expected. */
        if (debug_level > 0 || (idcode != 0 && idcode != 0xffffffff))
            fprintf (stderr, "%s: incompatible CPU detected, IDCODE=%08x\n", a->name, idcode);
        bitbang_ICSP_enable (a, 0);              // shut down target
        CloseHandle (a->devfd);
        free (a);
        return 0;
    }       

    /* Check status. */
    bitbang_send (a, 1, 1, 5, TAP_SW_MTAP, 0);    /* Send command. */
    bitbang_send (a, 1, 1, 5, MTAP_COMMAND, 0);   /* Send command. */
    bitbang_send (a, 0, 0, 8, MCHP_FLASH_ENABLE, 0);  /* Xfer data. */   // "This command requires a NOP to complete."
    bitbang_send (a, 0, 0, 8, MCHP_STATUS, 1);    /* Xfer data. */
    unsigned status = bitbang_recv (a);
    if (debug_level > 0)
        fprintf (stderr, "%s: [BB_5] status %04x\n", a->name, status);
    if ((status & ~MCHP_STATUS_DEVRST) !=
        (MCHP_STATUS_CPS | MCHP_STATUS_CFGRDY | MCHP_STATUS_FAEN)) {
        fprintf (stderr, "%s: [BB_6] invalid status = %04x\n", a->name, status);
        bitbang_ICSP_enable (a, 0);              // shut down target        
        CloseHandle (a->devfd);
        free (a);
        return 0;
    }
    printf ("      Adapter: %s\n", a->name);

    /* User functions. */
    a->adapter.close = bitbang_close;
    a->adapter.get_idcode = bitbang_get_idcode;
    a->adapter.load_executive = bitbang_load_executive;
    a->adapter.read_word = bitbang_read_word;
    a->adapter.read_data = bitbang_read_data;
    a->adapter.verify_data = bitbang_verify_data;
    a->adapter.erase_chip = bitbang_erase_chip;
    a->adapter.program_word = bitbang_program_word;
    a->adapter.program_row = bitbang_program_row;
    return &a->adapter;
}
